unit psvHost;
{$IFDEF VER140}
{$TYPEDADDRESS OFF} // Unit must be compiled without type-checked pointers.
{$WARN SYMBOL_PLATFORM OFF}
{$WRITEABLECONST ON}
{$ENDIF}

{$IFDEF VER150}
{$TYPEDADDRESS OFF} // Unit must be compiled without type-checked pointers.
{$WARN SYMBOL_PLATFORM OFF}
{$WRITEABLECONST ON}
{$VARPROPSETTER ON}
{$ENDIF}

{$R-,T-,H+,X+}

interface

uses Windows, Messages, ActiveX, SysUtils, Classes, Controls, Forms,
  Menus, Graphics, ComObj, AxCtrls, OleCtrls, ShlObj, ShellAPI, WinInet,
  Dialogs, {$IFNDEF VER130} Variants, {$ENDIF} Registry;

const
  ProcName:  PChar = 'DllRegisterServer';

type
  TTransferType  = (ttFileServer, ttFTPServer);
  TConnectType = (ctDirect, ctPreconfig, ctProxy);

  TRegProc = function : HResult; stdcall;
  TpsvApplet = class;
  TpsvAppletHost = class;

  TpsvServerOpt = class(TPersistent)
  private
   FOwner          : TpsvAppletHost;
   FUserName       : string;
   FPassword       : string;
   FServerFolder   : string;
   FServerName     : string;
   FAutoTransfer   : boolean;
   FTransferType   : TTransferType;
   FProxyServer    : string;
   FConnectType    : TConnectType;
  public
   constructor Create(AOwner : TpsvAppletHost);
   procedure Assign(Source: TPersistent); override;
  published
    property    TransferType : TTransferType read FTransferType write FTransferType;
    property    AutoTransfer : boolean read  FAutoTransfer write FAutoTransfer default false;
    property    ServerFolder : string read   FServerFolder write FServerFolder;
    property    ServerName   : string read   FServerName write FServerName;
    property    UserName     : string read   FUserName  write FUserName;
    property    Password     : string read   FPassword  write FPassword;
    property    ProxyServer  : string read   FProxyServer write FProxyServer;
    property    ConnectType  : TConnectType  read FConnectType write FConnectType;
  end;

  TpsvFileOpt = class(TPersistent)
  private
    FOwner         : TpsvAppletHost;
    FFileName      : string;
    FFolder        : string;
    FDefaultExt    : string;
    procedure SetFolder(Value : string);
    procedure SetFileName(Value : string);
    procedure SetDefaultExt(Value : string);
  public
    constructor Create(AOwner : TpsvAppletHost);
    procedure Assign(Source: TPersistent); override;
  published
    property    DefaultExt    : string read FDefaultExt write SetDefaultExt;
    property    FileName      : string read FFileName write SetFileName ;
    property    Folder        : string read FFolder write SetFolder;
  end;

  TpsvAppletHost = class(TWinControl,IDispatch)
  private
    FAutoRegister  : boolean;
    FAutoResize    : boolean;
    FServerOpt     : TpsvServerOpt;
    FFileOpt       : TpsvFileOpt;
    FApplet        : TpsvApplet;
    HI,FI          : HINTERNET;
    FAutoExtractGUID : boolean;
    FBeforeActivate   : TNotifyEvent;
    FAfterActivate    : TNotifyEvent;
    FBeforeDeactivate : TNotifyEvent;
    FAfterDeactivate  : TNotifyEvent;
    FBeforeDownload   : TNotifyEvent;
    FAfterDownload    : TNotifyEvent;
    FGUID          : TGUID;
    FGUIDStr       : string;
    FInactiveColor : TColor;
    FActive        : boolean;
    procedure SetActive(Value : boolean);
    procedure SetInactiveColor(Value : TColor);
    Procedure SetGUIDStr(Value : string);
    function  GetApplet : TpsvApplet;
    function  GetDefaultDispatch : IDispatch;
    procedure SetServerOpt(Value : TpsvServerOpt);
    procedure SetFileOpt(Value : TpsvFileOpt);
   protected
    { IDispatch }
    function GetTypeInfoCount(out Count: Integer): HResult; stdcall;
    function GetTypeInfo(Index, LocaleID: Integer; out TypeInfo): HResult; stdcall;
    function GetIDsOfNames(const IID: TGUID; Names: Pointer;
      NameCount, LocaleID: Integer; DispIDs: Pointer): HResult; stdcall;
    function Invoke(DispID: Integer; const IID: TGUID; LocaleID: Integer;
      Flags: Word; var Params; VarResult, ExcepInfo, ArgErr: Pointer): HResult; stdcall;
    procedure Loaded; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   RegisterApplet;
    procedure   GetAppletFromServer;
    procedure   GetAppletFromFTP;
    procedure   DownloadApplet;
    function    AppletRegistered : boolean;
    function    ExtractGUID : TGUID;
    procedure   BrowseProperties;
    property    Applet : TpsvApplet read GetApplet;
    property    DefaultDispatch : IDispatch read GetDefaultDispatch;
  published
    property    Ctl3D;
    property    Visible;
    property    BevelInner;
    property    BevelOuter;
    property    BevelKind;
    property    BevelWidth;
    property    BorderWidth;
    property    Active        : boolean read FActive write SetActive default false;
    property    InActiveColor : TColor read FInactiveColor write SetInactiveColor;
    property    GUID          : string read FGUIDStr  write SetGUIDStr;
    property    BeforeActivate: TNotifyEvent read FBeforeActivate write FBeforeActivate;
    property    AfterActivate : TNotifyEvent read FAfterActivate write FAfterActivate;
    property    BeforeDeactivate : TNotifyEvent read FBeforeDeactivate write FBeforeDeactivate;
    property    AfterDeactivate  : TNotifyEvent read FAfterDeactivate write FAfterDeactivate;
    property    AutoExtractGUID : boolean read FAutoExtractGUID write FAutoExtractGUID default true;
    property    ServerOptions : TpsvServerOpt Read FServerOpt write SetServerOpt;
    property    FileOptions   : TpsvFileOpt read FFileOpt write SetFileOpt;
    property    AutoResize    : boolean read FAutoResize write FAutoResize default true;
    property    AutoRegister  : boolean read FAutoRegister write FAutoRegister;
  end;


  TpsvApplet = class(TWinControl, IUnknown, IOleClientSite,
    IOleControlSite, IOleInPlaceSite, IOleInPlaceFrame, IDispatch,
    IPropertyNotifySink, ISimpleFrameSite)
  private
    FHost : TpsvAppletHost;
    FControlData: PControlData;
    FRefCount: Longint;
    FEventDispatch: TEventDispatch;
    FObjectData: HGlobal;
    FOleObject: IOleObject;
    FPersistStream: IPersistStreamInit;
    FOleControl: IOleControl;
    FControlDispatch: IDispatch;
    FPropBrowsing: IPerPropertyBrowsing;
    FOleInPlaceObject: IOleInPlaceObject;
    FOleInPlaceActiveObject: IOleInPlaceActiveObject;
    FPropConnection: Longint;
    FEventsConnection: Longint;
    FMiscStatus: Longint;
    FFonts: TList;
    FPictures: TList;
    FUpdatingPictures: Boolean;
    FUpdatingColor: Boolean;
    FUpdatingFont: Boolean;
    FUpdatingEnabled: Boolean;
    { TOleControl }
    procedure CreateControl;
    procedure CreateInstance;
    procedure DesignModified;
    procedure DestroyControl;
    procedure DestroyEnumPropDescs;
    function  GetMainMenu: TMainMenu;
    function  GetDefaultDispatch: IDispatch;
    procedure HookControlWndProc;
    procedure SetUIActive(Active: Boolean);
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure CMDocWindowActivate(var Message: TMessage); message CM_DOCWINDOWACTIVATE;
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMDialogKey(var Message: TMessage); message CM_DIALOGKEY;
    procedure CMUIActivate(var Message: TMessage); message CM_UIACTIVATE;
    procedure CMUIDeactivate(var Message: TMessage); message CM_UIDEACTIVATE;
  protected
    FEvents: Integer;
    { IUnknown }
    function QueryInterface(const IID: TGUID; out Obj): HResult; override;
    function _AddRef: Integer; stdcall;
    function _Release: Integer; stdcall;
    { IOleClientSite }
    function SaveObject: HResult; stdcall;
    function GetMoniker(dwAssign: Longint; dwWhichMoniker: Longint;
      out mk: IMoniker): HResult; stdcall;
    function GetContainer(out container: IOleContainer): HResult; stdcall;
    function ShowObject: HResult; stdcall;
    function OnShowWindow(fShow: BOOL): HResult; stdcall;
    function RequestNewObjectLayout: HResult; stdcall;
    { IOleControlSite }
    function OnControlInfoChanged: HResult; stdcall;
    function LockInPlaceActive(fLock: BOOL): HResult; stdcall;
    function GetExtendedControl(out disp: IDispatch): HResult; stdcall;
    function TransformCoords(var ptlHimetric: TPoint; var ptfContainer: TPointF;
      flags: Longint): HResult; stdcall;
    function IOleControlSite.TranslateAccelerator = OleControlSite_TranslateAccelerator;
    function OleControlSite_TranslateAccelerator(msg: PMsg;
      grfModifiers: Longint): HResult; stdcall;
    function OnFocus(fGotFocus: BOOL): HResult; stdcall;
    function ShowPropertyFrame: HResult; stdcall;
    { IOleWindow }
    function ContextSensitiveHelp(fEnterMode: BOOL): HResult; stdcall;
    { IOleInPlaceSite }
    function IOleInPlaceSite.GetWindow = OleInPlaceSite_GetWindow;
    function OleInPlaceSite_GetWindow(out wnd: HWnd): HResult; stdcall;
    function CanInPlaceActivate: HResult; stdcall;
    function OnInPlaceActivate: HResult; stdcall;
    function OnUIActivate: HResult; stdcall;
    function GetWindowContext(out frame: IOleInPlaceFrame;
      out doc: IOleInPlaceUIWindow; out rcPosRect: TRect;
      out rcClipRect: TRect; out frameInfo: TOleInPlaceFrameInfo): HResult;
      stdcall;
    function Scroll(scrollExtent: TPoint): HResult; stdcall;
    function OnUIDeactivate(fUndoable: BOOL): HResult; stdcall;
    function OnInPlaceDeactivate: HResult; stdcall;
    function DiscardUndoState: HResult; stdcall;
    function DeactivateAndUndo: HResult; stdcall;
    function OnPosRectChange(const rcPosRect: TRect): HResult; stdcall;
    { IOleInPlaceUIWindow }
    function GetBorder(out rectBorder: TRect): HResult; stdcall;
    function RequestBorderSpace(const borderwidths: TRect): HResult; stdcall;
    function SetBorderSpace(pborderwidths: PRect): HResult; stdcall;
    function SetActiveObject(const activeObject: IOleInPlaceActiveObject;
      pszObjName: POleStr): HResult; stdcall;
    { IOleInPlaceFrame }
    function IOleInPlaceFrame.GetWindow = OleInPlaceFrame_GetWindow;
    function OleInPlaceFrame_GetWindow(out wnd: HWnd): HResult; stdcall;
    function InsertMenus(hmenuShared: HMenu;
      var menuWidths: TOleMenuGroupWidths): HResult; stdcall;
    function SetMenu(hmenuShared: HMenu; holemenu: HMenu;
      hwndActiveObject: HWnd): HResult; stdcall;
    function RemoveMenus(hmenuShared: HMenu): HResult; stdcall;
    function SetStatusText(pszStatusText: POleStr): HResult; stdcall;
    function EnableModeless(fEnable: BOOL): HResult; stdcall;
    function IOleInPlaceFrame.TranslateAccelerator = OleInPlaceFrame_TranslateAccelerator;
    function OleInPlaceFrame_TranslateAccelerator(var msg: TMsg;
      wID: Word): HResult; stdcall;
    { IDispatch }
    function GetTypeInfoCount(out Count: Integer): HResult; stdcall;
    function GetTypeInfo(Index, LocaleID: Integer; out TypeInfo): HResult; stdcall;
    function GetIDsOfNames(const IID: TGUID; Names: Pointer;
      NameCount, LocaleID: Integer; DispIDs: Pointer): HResult; stdcall;
    function Invoke(DispID: Integer; const IID: TGUID; LocaleID: Integer;
      Flags: Word; var Params; VarResult, ExcepInfo, ArgErr: Pointer): HResult; stdcall;
    { ISimpleFrameSite }
    function PreMessageFilter(wnd: HWnd; msg, wp, lp: Integer;
      out res: Integer; out Cookie: Longint): HResult; stdcall;
    function PostMessageFilter(wnd: HWnd; msg, wp, lp: Integer;
      out res: Integer; Cookie: Longint): HResult; stdcall;
    { TOleControl }
    procedure CreateWnd; override;
    procedure DestroyWindowHandle; override;
    procedure GetProperty(Index: Integer; var Value: TVarData);
    procedure InitControlData; virtual;
    procedure InvokeMethod(const DispInfo; Result: Pointer);
    function  PaletteChanged(Foreground: Boolean): Boolean; override;
    procedure PictureChanged(Sender: TObject);
    procedure SetName(const Value: TComponentName); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure WndProc(var Message: TMessage); override;
    property ControlData: PControlData read FControlData write FControlData;
    { IPropertyNotifySink }
    function OnChanged(dispid: TDispID): HResult; virtual; stdcall;
    function OnRequestEdit(dispid: TDispID): HResult; virtual; stdcall;
    function GetHelpContext(Member: string; var HelpCtx: Integer;
      var HelpFile: string): Boolean;
    property    TabStop default True;
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   BrowseProperties;
    function    GetByteProp(Index: Integer): Byte;
    function    GetColorProp(Index: Integer): TColor;
    function    GetTColorProp(Index: Integer): TColor;
    function    GetCompProp(Index: Integer): Comp;
    function    GetCurrencyProp(Index: Integer): Currency;
    function    GetDoubleProp(Index: Integer): Double;
    function    GetIDispatchProp(Index: Integer): IDispatch;
    function    GetIntegerProp(Index: Integer): Integer;
    function    GetIUnknownProp(Index: Integer): IUnknown;
    function    GetWordBoolProp(Index: Integer): WordBool;
    function    GetTDateTimeProp(Index: Integer): TDateTime;
    function    GetTFontProp(Index: Integer): TFont;
    function    GetOleBoolProp(Index: Integer): TOleBool;
    function    GetOleDateProp(Index: Integer): TOleDate;
    function    GetOleEnumProp(Index: Integer): TOleEnum;
    function    GetTOleEnumProp(Index: Integer): TOleEnum;
    function    GetOleVariantProp(Index: Integer): OleVariant;
    function    GetTPictureProp(Index: Integer): TPicture;
    function    GetShortIntProp(Index: Integer): ShortInt;
    function    GetSingleProp(Index: Integer): Single;
    function    GetSmallintProp(Index: Integer): Smallint;
    function    GetStringProp(Index: Integer): string;
    function    GetVariantProp(Index: Integer): Variant;
    function    GetWideStringProp(Index: Integer): WideString;
    function    GetWordProp(Index: Integer): Word;
    procedure   DefaultHandler(var Message); override;
    procedure   DoObjectVerb(Verb: Integer);
    procedure   SetByteProp(Index: Integer; Value: Byte);
    procedure   SetColorProp(Index: Integer; Value: TColor);
    procedure   SetTColorProp(Index: Integer; Value: TColor);
    procedure   SetCompProp(Index: Integer; const Value: Comp);
    procedure   SetCurrencyProp(Index: Integer; const Value: Currency);
    procedure   SetDoubleProp(Index: Integer; const Value: Double);
    procedure   SetIDispatchProp(Index: Integer; const Value: IDispatch);
    procedure   SetIntegerProp(Index: Integer; Value: Integer);
    procedure   SetIUnknownProp(Index: Integer; const Value: IUnknown);
    procedure   SetWordBoolProp(Index: Integer; Value: WordBool);
    procedure   SetTDateTimeProp(Index: Integer; const Value: TDateTime);
    procedure   SetTFontProp(Index: Integer; Value: TFont);
    procedure   SetOleBoolProp(Index: Integer; Value: TOleBool);
    procedure   SetOleDateProp(Index: Integer; const Value: TOleDate);
    procedure   SetOleEnumProp(Index: Integer; Value: TOleEnum);
    procedure   SetTOleEnumProp(Index: Integer; Value: TOleEnum);
    procedure   SetOleVariantProp(Index: Integer; const Value: OleVariant);
    procedure   SetTPictureProp(Index: Integer;  Value: TPicture);
    procedure   SetProperty(Index: Integer; const Value: TVarData);
    procedure   SetShortIntProp(Index: Integer; Value: Shortint);
    procedure   SetSingleProp(Index: Integer; const Value: Single);
    procedure   SetSmallintProp(Index: Integer; Value: Smallint);
    procedure   SetStringProp(Index: Integer; const Value: string);
    procedure   SetVariantProp(Index: Integer; const Value: Variant);
    procedure   SetWideStringProp(Index: Integer; const Value: WideString);
    procedure   SetWordProp(Index: Integer; Value: Word);
    procedure   GetObjectVerbs(List: TStrings);
    function    GetPropDisplayString(DispID: Integer): string;
    procedure   GetPropDisplayStrings(DispID: Integer; List: TStrings);
    function    IsCustomProperty(DispID: Integer): Boolean;
    function    IsPropPageProperty(DispID: Integer): Boolean;
    procedure   SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure   SetPropDisplayString(DispID: Integer; const Value: string);
    property    PerPropBrowsing: IPerPropertyBrowsing read FPropBrowsing;
    property    DefaultDispatch: IDispatch read GetDefaultDispatch;
  end;


function  FontToOleFont(Font: TFont): Variant;
procedure OleFontToFont(const OleFont: Variant; Font: TFont);
procedure Exchange(var A,B); register;
procedure Register;

implementation

uses OleConst;

const
  OCM_BASE = $2000;

{ Control flags }

const
  cfBackColor = $00000001;
  cfForeColor = $00000002;
  cfFont      = $00000004;
  cfEnabled   = $00000008;
  cfCaption   = $00000010;
  cfText      = $00000020;

const
  MaxDispArgs = 32;

type
  PDispInfo = ^TDispInfo;
  TDispInfo = packed record
    DispID: TDispID;
    ResType: Byte;
    CallDesc: TCallDesc;
  end;

  TArgKind = (akDWord, akSingle, akDouble);

  PEventArg = ^TEventArg;
  TEventArg = record
    Kind: TArgKind;
    Data: array[0..1] of Integer;
  end;

  TEventInfo = record
    Method: TMethod;
    Sender: TObject;
    ArgCount: Integer;
    Args: array[0..MaxDispArgs - 1] of TEventArg;
  end;


function StringToVarOleStr(const S: string): Variant;
begin
  VarClear(Result);
  TVarData(Result).VOleStr := StringToOleStr(S);
  TVarData(Result).VType := varOleStr;
end;

procedure Exchange(var A,B); register;
asm
  MOV   ECX, [EDX]
  XCHG  ECX, [EAX]
  MOV   [EDX], ECX
end;

procedure INetCheck(b: LongBool);
var
   ErrNum   : integer;
   TmpErr   : DWORD;
   Len      : DWORD;
   Buf      : array[0..4096] of char;
   s        : string;
begin
   if not b then begin
      ErrNum := GetLastError;
      case ErrNum of
         ERROR_INTERNET_EXTENDED_ERROR : begin
            Len := sizeof(Buf);
            InternetGetLastResponseInfo(TmpErr, Buf, Len);
            Buf[Len] := #0;
            s := Format('Win32 Error %d:'#13#10#13#10'%s', [ErrNum, Buf]);
            end;

         { Internet API error returns }
         ERROR_INTERNET_TIMEOUT                      : s := 'INTERNET TIMEOUT ';
         ERROR_INTERNET_INTERNAL_ERROR               : s := 'INTERNET INTERNAL ERROR';
         ERROR_INTERNET_INVALID_URL                  : s := 'INTERNET INVALID URL ';
         ERROR_INTERNET_UNRECOGNIZED_SCHEME          : s := 'INTERNET UNRECOGNIZED SCHEME ';
         ERROR_INTERNET_NAME_NOT_RESOLVED            : s := 'INTERNET NAME NOT RESOLVED ';
         ERROR_INTERNET_PROTOCOL_NOT_FOUND           : s := 'INTERNET PROTOCOL NOT FOUND';
         ERROR_INTERNET_INVALID_OPTION               : s := 'INTERNET INVALID OPTION';
         ERROR_INTERNET_BAD_OPTION_LENGTH            : s := 'INTERNET BAD OPTION LENGTH ';
         ERROR_INTERNET_OPTION_NOT_SETTABLE          : s := 'INTERNET OPTION NOT SETTABLE ';
         ERROR_INTERNET_SHUTDOWN                     : s := 'INTERNET SHUTDOWN';
         ERROR_INTERNET_INCORRECT_USER_NAME          : s := 'INTERNET INCORRECT USER NAME ';
         ERROR_INTERNET_INCORRECT_PASSWORD           : s := 'INTERNET INCORRECT PASSWORD';
         ERROR_INTERNET_LOGIN_FAILURE                : s := 'INTERNET LOGIN FAILURE ';
         ERROR_INTERNET_INVALID_OPERATION            : s := 'INTERNET INVALID OPERATION ';
         ERROR_INTERNET_OPERATION_CANCELLED          : s := 'INTERNET OPERATION CANCELLED ';
         ERROR_INTERNET_INCORRECT_HANDLE_TYPE        : s := 'INTERNET INCORRECT HANDLE TYPE ';
         ERROR_INTERNET_INCORRECT_HANDLE_STATE       : s := 'INTERNET INCORRECT HANDLE STATE';
         ERROR_INTERNET_NOT_PROXY_REQUEST            : s := 'INTERNET NOT PROXY REQUEST ';
         ERROR_INTERNET_REGISTRY_VALUE_NOT_FOUND     : s := 'INTERNET REGISTRY VALUE NOT FOUND';
         ERROR_INTERNET_BAD_REGISTRY_PARAMETER       : s := 'INTERNET BAD REGISTRY PARAMETER';
         ERROR_INTERNET_NO_DIRECT_ACCESS             : s := 'INTERNET NO DIRECT ACCESS';
         ERROR_INTERNET_NO_CONTEXT                   : s := 'INTERNET NO CONTEXT';
         ERROR_INTERNET_NO_CALLBACK                  : s := 'INTERNET NO CALLBACK ';
         ERROR_INTERNET_REQUEST_PENDING              : s := 'INTERNET REQUEST PENDING ';
         ERROR_INTERNET_INCORRECT_FORMAT             : s := 'INTERNET INCORRECT FORMAT';
         ERROR_INTERNET_ITEM_NOT_FOUND               : s := 'INTERNET ITEM NOT FOUND';
         ERROR_INTERNET_CANNOT_CONNECT               : s := 'INTERNET CANNOT CONNECT';
         ERROR_INTERNET_CONNECTION_ABORTED           : s := 'INTERNET CONNECTION ABORTED';
         ERROR_INTERNET_CONNECTION_RESET             : s := 'INTERNET CONNECTION RESET';
         ERROR_INTERNET_FORCE_RETRY                  : s := 'INTERNET FORCE RETRY ';
         ERROR_INTERNET_INVALID_PROXY_REQUEST        : s := 'INTERNET INVALID PROXY REQUEST ';

         ERROR_INTERNET_HANDLE_EXISTS                : s := 'INTERNET HANDLE_EXISTS ';
         ERROR_INTERNET_SEC_CERT_DATE_INVALID        : s := 'INTERNET SEC CERT DATE INVALID ';
         ERROR_INTERNET_SEC_CERT_CN_INVALID          : s := 'INTERNET SEC CERT CN INVALID ';
         ERROR_INTERNET_HTTP_TO_HTTPS_ON_REDIR       : s := 'INTERNET HTTP TO HTTPS ON REDIR';
         ERROR_INTERNET_HTTPS_TO_HTTP_ON_REDIR       : s := 'INTERNET HTTPS TO HTTP ON REDIR';
         ERROR_INTERNET_MIXED_SECURITY               : s := 'INTERNET MIXED SECURITY';
         ERROR_INTERNET_CHG_POST_IS_NON_SECURE       : s := 'INTERNET CHG POST IS NON SECURE';
         ERROR_INTERNET_POST_IS_NON_SECURE           : s := 'INTERNET POST IS NON SECURE';

         { FTP API errors }
         ERROR_FTP_TRANSFER_IN_PROGRESS              : s := 'FTP TRANSFER IN PROGRESS ';
         ERROR_FTP_DROPPED                           : s := 'FTP DROPPED    ';

         { gopher API errors }
         ERROR_GOPHER_PROTOCOL_ERROR                 : s := 'GOPHER_PROTOCOL_ERROR';
         ERROR_GOPHER_NOT_FILE                       : s := 'GOPHER_NOT_FILE';
         ERROR_GOPHER_DATA_ERROR                     : s := 'GOPHER_DATA_ERROR';
         ERROR_GOPHER_END_OF_DATA                    : s := 'GOPHER_END_OF_DATA ';
         ERROR_GOPHER_INVALID_LOCATOR                : s := 'GOPHER_INVALID_LOCATOR ';
         ERROR_GOPHER_INCORRECT_LOCATOR_TYPE         : s := 'GOPHER_INCORRECT_LOCATOR_TYPE';
         ERROR_GOPHER_NOT_GOPHER_PLUS                : s := 'GOPHER_NOT_GOPHER_PLUS ';
         ERROR_GOPHER_ATTRIBUTE_NOT_FOUND            : s := 'GOPHER_ATTRIBUTE_NOT_FOUND ';
         ERROR_GOPHER_UNKNOWN_LOCATOR                : s := 'GOPHER_UNKNOWN_LOCATOR ';

         { HTTP API errors }
         ERROR_HTTP_HEADER_NOT_FOUND                 : s := 'HTTP_HEADER_NOT_FOUND';
         ERROR_HTTP_DOWNLEVEL_SERVER                 : s := 'HTTP_DOWNLEVEL_SERVER';
         ERROR_HTTP_INVALID_SERVER_RESPONSE          : s := 'HTTP_INVALID_SERVER_RESPONSE ';
         ERROR_HTTP_INVALID_HEADER                   : s := 'HTTP_INVALID_HEADER';
         ERROR_HTTP_INVALID_QUERY_REQUEST            : s := 'HTTP_INVALID_QUERY_REQUEST ';
         ERROR_HTTP_HEADER_ALREADY_EXISTS            : s := 'HTTP_HEADER_ALREADY_EXISTS ';
         ERROR_HTTP_REDIRECT_FAILED                  : s := 'HTTP_REDIRECT_FAILED ';

         else RaiseLastWin32Error;
      end;

      raise Exception.Create(s);
   end;
end;

{ TpsvAppletHost }

const
  cdForceSetClientSite = 1;
  cdDeferSetClientSite = 2;

constructor TpsvAppletHost.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FFileOpt := TpsvFileOpt.Create(Self);
  FServerOpt := TpsvServerOpt.Create(Self);
  ControlStyle := ControlStyle + [csOpaque];
  HI := NIL;
  FI := NIL;
  FActive := False;
  FAutoExtractGUID := true;
  Width   := 100;
  Height  := 100;
  FInactiveColor   := clBlack;
  Color := FInactiveColor;
  FGUIDStr    := EmptyStr;
  FGUID       := GUID_NULL;
  FAutoResize := true;
end;

procedure TpsvAppletHost.Loaded;
begin
 inherited;
 if FActive then
  begin
    FActive := False;
    SetActive(True);
  end;
end;

function TpsvAppletHost.ExtractGUID : TGUID;
var TypeLib : ITypeLIB;
    I : Integer;
    TypeInfo : ActiveX.ITypeInfo;
    TypeAttr : PTypeAttr;
    TK : ActiveX.TTypeKIND;
begin
  Result := GUID_NULL;
  if FFileOpt.FileName = EmptyStr then Exit;
  OleCheck(LoadTypeLib(PWideChar(WideString(FFileOpt.Folder+FFileOpt.FileName)), TypeLib));
  For I := 0 to TypeLib.GetTypeInfoCount -1 do
  begin
    TypeLib.GetTypeInfoType(I,TK);
    if TK = TKIND_COCLASS then
     begin
        TypeLib.GetTypeInfo(I,TypeInfo);
        TypeInfo.GetTypeAttr(TypeAttr);
        Result := TypeAttr.guid;
        TypeInfo.ReleaseTypeAttr(TypeAttr);
        Exit;
     end;
  end;
end;

function  TpsvAppletHost.GetApplet : TpsvApplet;
begin
  Result := nil;
  if (csDesigning in ComponentState) then
     Exit;
  if not Active then Exit;
  Result := FApplet;
end;

function TpsvAppletHost.GetDefaultDispatch : IDispatch;
begin
  Result := nil;
  if (csDesigning in ComponentState) then
     Exit;
  if not Active then Exit;
  Result := FApplet.DefaultDispatch;
end;

Procedure  TpsvAppletHost.SetGUIDStr(Value : string);
begin
  If Value <> FGUIDStr then
   begin
     FGUIDStr := Value;
     FGUID := StringToGUID(FGUIDStr);
   end;
end;

procedure TpsvAppletHost.SetServerOpt(Value : TpsvServerOpt);
begin
  FServerOpt.Assign(Value);
end;

procedure TpsvAppletHost.SetFileOpt(Value : TpsvFileOpt);
begin
  FFileOpt.Assign(Value);
end;

procedure  TpsvAppletHost.DownloadApplet;
begin
  if FServerOpt.TransferType = ttFileServer then
    GetAppletFromServer else GetAppletFromFTP;
end;

procedure TpsvAppletHost.BrowseProperties;
begin
 if FActive then
   FApplet.BrowseProperties;
end;


procedure TpsvAppletHost.SetActive(Value : boolean);
begin

If (FActive = Value) then Exit;
if csLoading in ComponentState then
   FActive := Value else
 begin

   if not FAutoExtractGUID then
    begin
      IF (FGUIDStr = EmptyStr) then
        Raise Exception.Create('GUID Unknown');
    end;

    if FFileOpt.FileName = EmptyStr then
       raise Exception.Create('File name missing');

  FActive := Value;
  if not FActive then
   begin
      if Assigned(FBeforeDeactivate) then
        FBeforeDeactivate(Self);
       FApplet.Free;
       FApplet := nil;
       Color := FInactiveColor;
       RecreateWnd;
      if Assigned(FAfterDeactivate) then
        FAfterDeactivate(Self);
   end
     else
       begin
         if not ( csDesigning in ComponentState) then
          begin
            if not FileExists(FFileopt.Folder + FFileopt.FileName) then
              begin
                if FServerOpt.AutoTransfer then
                   DownloadApplet;
              end;
          end;

         if not FileExists(FFileOpt.Folder + FFileOpt.FileName) then
          begin
            FActive := False;
            Raise Exception.CreateFmt('File %s not exists in %s',[FFileOpt.FileName, FFileOpt.Folder]);
          end;

         if FGUIDStr = EmptyStr then
          begin
            if FAutoExtractGUID then
              begin
                 FGUID := ExtractGUID;
                 FGUIDStr := GUIDToString(FGUID);
              end;
          end;

         if not AppletRegistered then
         begin
           if FAutoRegister then
            RegisterApplet else
            begin
              FActive := false;
              raise Exception.Create('Applet not registered');
            end;  
         end;

         if Assigned(FBeforeActivate) then
           FBeforeActivate(Self);
         FApplet := TpsvApplet.Create(Self);
         if FAutoResize then
          begin
           Width := FApplet.Width;
           Height := FApplet.Height;
         end;
         FApplet.Parent := Self;
         FApplet.Align := alClient;
         if Assigned(FAfterActivate) then
           FAfterActivate(Self);
       end;
  end;
end;

destructor TpsvAppletHost.Destroy;
begin
  if FActive then
    SetActive(False);
  FServerOpt.Free;
  FFileOpt.Free;
  inherited Destroy;
end;

function  TpsvAppletHost.AppletRegistered : boolean;
var Reg : TRegistry;
begin
  Reg := TRegistry.Create;
  Reg.RootKey := HKEY_CLASSES_ROOT;
  Result := Reg.OpenKey('CLSID\' + GuidToString(FGUID), False);
  Reg.CloseKey;
  Reg.Free;
end;

procedure TpsvAppletHost.SetInactiveColor(Value : TColor);
begin
  If FInactiveColor <> Value then
    begin
      FInactiveColor := Value;
      if not FActive then
        begin
          Color := FInactiveColor;
          RecreateWnd;
       end;
   end;
end;

procedure  TpsvAppletHost.GetAppletFromFTP;
begin
 if Assigned(FBeforeDownload) then
  FBeforeDownload(Self);
 try
  case FServerOpt.ConnectType of
        ctDirect : HI := InternetOpen('Microsoft Internet Explorer', INTERNET_OPEN_TYPE_DIRECT, nil, nil, 0);
        ctProxy  : HI := InternetOpen('Microsoft Internet Explorer', INTERNET_OPEN_TYPE_PROXY, PChar(FServerOpt.ProxyServer), nil, 0);
        ctPreconfig  : HI := InternetOpen('Microsoft Internet Explorer', INTERNET_OPEN_TYPE_PRECONFIG, nil, nil, 0);
  end;
  INetCheck(LongBool(hi));
  FI := InternetConnect(HI, PChar(FServeropt.ServerName), INTERNET_INVALID_PORT_NUMBER,
          PChar(FServerOpt.UserName), PChar(FServerOpt.Password), INTERNET_SERVICE_FTP, 0, 0);

  INetCheck(LongBool(fi));
  INetCheck(FtpGetFile(FI, PChar(FServerOpt.ServerFolder+FFileOpt.FileName),
   PChar(FFileOpt.Folder+FFileOpt.FileName), False, INTERNET_FLAG_RELOAD,
                       FTP_TRANSFER_TYPE_BINARY, DWORD(Self)));
 finally

  INetCheck(InternetCloseHandle(FI));
  INetCheck(InternetCloseHandle(HI));
 end;
 if Assigned(FAfterDownload) then
  FAfterDownload(Self);
end;

procedure  TpsvAppletHost.GetAppletFromServer;
var
  ShFileOpStruct: TShFileOpStruct;
begin
  if Assigned(FBeforeDownload) then
   FBeforeDownload(Self);
  if FServerOpt.ServerFolder[Length(FServerOpt.ServerFolder)] <> '\' then
    FServerOpt.ServerFolder := FServerOpt.ServerFolder + '\';
  ShFileOpStruct.Wnd   := 0;
  ShFileOpStruct.wFunc := FO_COPY;
  ShFileOpStruct.pFrom :=PChar(FServerOpt.ServerFolder+FFileOpt.FileName+#0);
  ShFileOpStruct.pTo   :=pchar(FFileOpt.Folder+#0);
  ShFileOpStruct.fFlags := FOF_ALLOWUNDO or FOF_NOCONFIRMATION;
  if ( ShFileOperation(ShFileOpStruct) <> 0) then
   raise Exception.Create('Error while copying file from server');
 if assigned(FAfterDownload) then
  FAfterDownload(Self);
end;


procedure TpsvAppletHost.RegisterApplet;
var
  LibHandle: THandle;
  RegProc: TRegProc;
begin
  IF FFileOpt.FileName = EmptyStr then
    raise Exception.Create('Unknown file name');
  LibHandle := LoadLibrary(PChar(FFileOpt.Folder + FFileOpt.FileName));
  if LibHandle = 0 then raise Exception.CreateFmt('Failed to load "%s"', [FFileOpt.FileName]);
  try
    @RegProc := GetProcAddress(LibHandle, ProcName);
    if @RegProc = Nil then
      raise Exception.CreateFmt('%s procedure not found in "%s"', [ProcName,FFileOpt.FileName]);
    if RegProc <> 0 then
      raise Exception.CreateFmt('Call to %s failed in "%s"', [ProcName,FFileOpt.FileName]);
  finally
    FreeLibrary(LibHandle);
  end;
end;

function TpsvAppletHost.GetTypeInfoCount(out Count: Integer): HResult;
begin
  Count := 0;
  Result := S_OK;
end;

function TpsvAppletHost.GetTypeInfo(Index, LocaleID: Integer;
  out TypeInfo): HResult;
begin
  Pointer(TypeInfo) := nil;
  Result := E_NOTIMPL;
end;

function TpsvAppletHost.GetIDsOfNames(const IID: TGUID; Names: Pointer;
  NameCount, LocaleID: Integer; DispIDs: Pointer): HResult;
begin
  Result := E_NOTIMPL;
end;

function TpsvAppletHost.Invoke(DispID: Integer; const IID: TGUID;
  LocaleID: Integer; Flags: Word; var Params;
  VarResult, ExcepInfo, ArgErr: Pointer): HResult;
var
  F: TFont;
begin
  if (Flags and DISPATCH_PROPERTYGET <> 0) and (VarResult <> nil) then
  begin
    Result := S_OK;
    case DispID of
      DISPID_AMBIENT_BACKCOLOR:
        PVariant(VarResult)^ := Color;
      DISPID_AMBIENT_DISPLAYNAME:
        PVariant(VarResult)^ := StringToVarOleStr(Name);
      DISPID_AMBIENT_FONT:
      begin
        if (Parent <> nil) and ParentFont then
          F := TpsvAppletHost(Parent).Font
        else
          F := Font;
        PVariant(VarResult)^ := FontToOleFont(F);
      end;
      DISPID_AMBIENT_FORECOLOR:
        PVariant(VarResult)^ := Font.Color;
      DISPID_AMBIENT_LOCALEID:
        PVariant(VarResult)^ := Integer(GetUserDefaultLCID);
      DISPID_AMBIENT_MESSAGEREFLECT:
        PVariant(VarResult)^ := True;
      DISPID_AMBIENT_USERMODE:
        PVariant(VarResult)^ := not (csDesigning in ComponentState);
      DISPID_AMBIENT_UIDEAD:
        PVariant(VarResult)^ := csDesigning in ComponentState;
      DISPID_AMBIENT_SHOWGRABHANDLES:
        PVariant(VarResult)^ := False;
      DISPID_AMBIENT_SHOWHATCHING:
        PVariant(VarResult)^ := False;
      DISPID_AMBIENT_SUPPORTSMNEMONICS:
        PVariant(VarResult)^ := True;
      DISPID_AMBIENT_AUTOCLIP:
        PVariant(VarResult)^ := True;
    else
      Result := DISP_E_MEMBERNOTFOUND;
    end;
  end else
    Result := DISP_E_MEMBERNOTFOUND;
end;


constructor TpsvApplet.Create(AOwner: TComponent);
var
  I: Integer;
begin
  inherited Create(AOwner);
  FHost := TpsvAppletHost(AOwner);
  Include(FComponentStyle, csCheckPropAvail);
  InitControlData;
  Inc(FControlData^.InstanceCount);
  if FControlData^.FontCount > 0 then
  begin
    FFonts := TList.Create;
    FFonts.Count := FControlData^.FontCount;
    for I := 0 to FFonts.Count-1 do
      FFonts[I] := TFont.Create;
  end;
  if FControlData^.PictureCount > 0 then
  begin
    FPictures := TList.Create;
    FPictures.Count := FControlData^.PictureCount;
    for I := 0 to FPictures.Count-1 do
    begin
      FPictures[I] := TPicture.Create;
      TPicture(FPictures[I]).OnChange := PictureChanged;
    end;
  end;
  CreateInstance;
  OleCheck(FOleObject.GetMiscStatus(DVASPECT_CONTENT, FMiscStatus));
  if (FControlData^.Reserved and cdDeferSetClientSite) = 0 then
    if ((FMiscStatus and OLEMISC_SETCLIENTSITEFIRST) <> 0) or
      ((FControlData^.Reserved and cdForceSetClientSite) <> 0) then
      OleCheck(FOleObject.SetClientSite(Self));
  OleCheck(FOleObject.QueryInterface(IPersistStreamInit, FPersistStream));
  if FMiscStatus and OLEMISC_INVISIBLEATRUNTIME <> 0 then
    Visible := False;
  if FMiscStatus and OLEMISC_SIMPLEFRAME <> 0 then
    ControlStyle := [csAcceptsControls, csDoubleClicks, csNoStdEvents] else
    ControlStyle := [csDoubleClicks, csNoStdEvents];
  TabStop := FMiscStatus and (OLEMISC_ACTSLIKELABEL or
    OLEMISC_NOUIACTIVATE) = 0;
  OleCheck(RequestNewObjectLayout);
end;

destructor TpsvApplet.Destroy;

  procedure FreeList(var L: TList);
  var
    I: Integer;
  begin
    if L <> nil then
    begin
      for I := 0 to L.Count-1 do
        TObject(L[I]).Free;
      L.Free;
      L := nil;
    end;
  end;

begin
  SetUIActive(False);
  if FOleObject <> nil then FOleObject.Close(OLECLOSE_NOSAVE);
  DestroyControl;
  FPersistStream := nil;
  if FOleObject <> nil then FOleObject.SetClientSite(nil);
  FOleObject := nil;
  FEventDispatch.Free;
  FreeList(FFonts);
  FreeList(FPictures);
  Dec(FControlData^.InstanceCount);
  if FControlData^.InstanceCount = 0 then DestroyEnumPropDescs;
  inherited Destroy;
end;

procedure TpsvApplet.BrowseProperties;
begin
  DoObjectVerb(OLEIVERB_PROPERTIES);
end;

procedure TpsvApplet.CreateControl;
var
  Stream: IStream;
  CS: IOleClientSite;
  X: Integer;
begin
  if FOleControl = nil then
    try
      try  // work around ATL bug
        X := FOleObject.GetClientSite(CS);
      except
        X := -1;
      end;
      if (X <> 0) or (CS = nil) then
        OleCheck(FOleObject.SetClientSite(Self));
      if FObjectData = 0 then OleCheck(FPersistStream.InitNew) else
      begin
        OleCheck(CreateStreamOnHGlobal(FObjectData, False, Stream));
        OleCheck(FPersistStream.Load(Stream));
      end;
      OleCheck(FOleObject.QueryInterface(IOleControl, FOleControl));
      OleCheck(FOleObject.QueryInterface(IDispatch, FControlDispatch));
      FOleObject.QueryInterface(IPerPropertyBrowsing, FPropBrowsing);
      InterfaceConnect(FOleObject, IPropertyNotifySink,
        Self, FPropConnection);
      InterfaceConnect(FOleObject, FControlData^.EventIID,
        FEventDispatch, FEventsConnection);
      if FControlData^.Flags and cfBackColor <> 0 then
        OnChanged(DISPID_BACKCOLOR);
      if FControlData^.Flags and cfEnabled <> 0 then
        OnChanged(DISPID_ENABLED);
      if FControlData^.Flags and cfFont <> 0 then
        OnChanged(DISPID_FONT);
      if FControlData^.Flags and cfForeColor <> 0 then
        OnChanged(DISPID_FORECOLOR);
      FOleControl.OnAmbientPropertyChange(DISPID_UNKNOWN);
      RequestNewObjectLayout;
    except
      DestroyControl;
      raise;
    end;
end;


procedure TpsvApplet.CreateInstance;
var
  ClassFactory2: IClassFactory2;
  LicKeyStr: WideString;

  procedure LicenseCheck(Status: HResult; const Ident: string);
  begin
    if Status = CLASS_E_NOTLICENSED then
      raise EOleError.CreateFmt(Ident, [ClassName]);
    OleCheck(Status);
  end;

begin
  if not (csDesigning in ComponentState) and
    (FControlData^.LicenseKey <> nil) then
  begin
    OleCheck(CoGetClassObject(FControlData^.ClassID, CLSCTX_INPROC_SERVER or
      CLSCTX_LOCAL_SERVER, nil, IClassFactory2, ClassFactory2));
    LicKeyStr := PWideChar(FControlData^.LicenseKey);
    LicenseCheck(ClassFactory2.CreateInstanceLic(nil, nil, IOleObject,
      LicKeyStr, FOleObject), SInvalidLicense);
  end else
    LicenseCheck(CoCreateInstance(FControlData^.ClassID, nil,
      CLSCTX_INPROC_SERVER or CLSCTX_LOCAL_SERVER, IOleObject,
      FOleObject), SNotLicensed);
end;


procedure TpsvApplet.CreateWnd;
begin
  CreateControl;
  if FMiscStatus and OLEMISC_INVISIBLEATRUNTIME = 0 then
  begin
    FOleObject.DoVerb(OLEIVERB_INPLACEACTIVATE, nil, Self, 0,
      GetParentHandle, BoundsRect);
    if FOleInPlaceObject = nil then
      raise EOleError.CreateRes(@SCannotActivate);
    HookControlWndProc;
    if not Visible and IsWindowVisible(Handle) then
      ShowWindow(Handle, SW_HIDE);
  end else
    inherited CreateWnd;
end;

procedure TpsvApplet.DefaultHandler(var Message);
begin
  if HandleAllocated then
    with TMessage(Message) do
    begin
      if (Msg >= CN_BASE) and (Msg < CN_BASE + WM_USER) then
        Msg := Msg - (CN_BASE - OCM_BASE);
      if FMiscStatus and OLEMISC_SIMPLEFRAME = 0 then
      begin
        Result := CallWindowProc(DefWndProc, Handle, Msg, WParam, LParam);
        Exit;
      end;
    end;
  inherited DefaultHandler(Message);
end;


procedure TpsvApplet.DesignModified;
var
  Form: TCustomForm;
begin
  Form := GetParentForm(Self);
  if (Form <> nil) and (Form.Designer <> nil) then Form.Designer.Modified;
end;

procedure TpsvApplet.DestroyControl;
begin
  InterfaceDisconnect(FOleObject, FControlData^.EventIID, FEventsConnection);
  InterfaceDisconnect(FOleObject, IPropertyNotifySink, FPropConnection);
  FPropBrowsing := nil;
  FControlDispatch := nil;
  FOleControl := nil;
end;

procedure TpsvApplet.DestroyEnumPropDescs;
var
  I: Integer;
begin
  with FControlData^ do
    if EnumPropDescs <> nil then
    begin
      for I := 0 to EnumPropDescs.Count - 1 do
        TEnumPropDesc(EnumPropDescs[I]).Free;
      EnumPropDescs.Free;
      EnumPropDescs := nil;
    end;
end;


procedure TpsvApplet.DestroyWindowHandle;
begin
  if FMiscStatus and OLEMISC_INVISIBLEATRUNTIME = 0 then
  begin
    SetWindowLong(WindowHandle, GWL_WNDPROC, Longint(DefWndProc));
    if FOleObject <> nil then FOleObject.Close(OLECLOSE_NOSAVE);
    WindowHandle := 0;
  end else
    inherited DestroyWindowHandle;
end;

procedure TpsvApplet.DoObjectVerb(Verb: Integer);
var
  ActiveWindow: HWnd;
  WindowList: Pointer;
begin
  CreateControl;
  ActiveWindow := GetActiveWindow;
  WindowList := DisableTaskWindows(0);
  try
    OleCheck(FOleObject.DoVerb(Verb, nil, Self, 0,
      GetParentHandle, BoundsRect));
  finally
    EnableTaskWindows(WindowList);
    SetActiveWindow(ActiveWindow);
    Windows.SetFocus(ActiveWindow);
  end;
  if FPersistStream.IsDirty <> S_FALSE then DesignModified;
end;

function TpsvApplet.GetByteProp(Index: Integer): Byte;
begin
  Result := GetIntegerProp(Index);
end;

function TpsvApplet.GetColorProp(Index: Integer): TColor;
begin
  Result := GetIntegerProp(Index);
end;

function TpsvApplet.GetTColorProp(Index: Integer): TColor;
begin
  Result := GetIntegerProp(Index);
end;

function TpsvApplet.GetCompProp(Index: Integer): Comp;
begin
  Result := GetDoubleProp(Index);
end;

function TpsvApplet.GetCurrencyProp(Index: Integer): Currency;
var
  Temp: TVarData;
begin
  GetProperty(Index, Temp);
  Result := Temp.VCurrency;
end;

function TpsvApplet.GetDoubleProp(Index: Integer): Double;
var
  Temp: TVarData;
begin
  GetProperty(Index, Temp);
  Result := Temp.VDouble;
end;


{ TpsvApplet.GetHelpContext:  Fetch the help file name and help context
  id of the given member (property, event, or method) of the Ole Control from
  the control's ITypeInfo interfaces.  GetHelpContext returns False if
  the member name is not found in the control's ITypeInfo.
  To obtain a help context for the entire control class, pass an empty
  string as the Member name.  }

function TpsvApplet.GetHelpContext(Member: string;
  var HelpCtx: Integer; var HelpFile: string): Boolean;
var
  TypeInfo: ITypeInfo;
  HlpFile: TBStr;
  ImplTypes, MemberID: Integer;
  TypeAttr: PTypeAttr;

  function Find(const MemberStr: string; var TypeInfo: ITypeInfo): Boolean;
  var
    Code: HResult;
    I, Flags: Integer;
    RefType: HRefType;
    Name: TBStr;
    Temp: ITypeInfo;
  begin
    Result := False;
    Name := StringToOleStr(Member);
    try
      I := 0;
      while (I < ImplTypes) do
      begin
        OleCheck(TypeInfo.GetImplTypeFlags(I, Flags));
        if Flags and (IMPLTYPEFLAG_FDEFAULT or IMPLTYPEFLAG_FSOURCE) <> 0 then
        begin
          OleCheck(TypeInfo.GetRefTypeOfImplType(I, RefType));
          OleCheck(TypeInfo.GetRefTypeInfo(RefType, Temp));
          Code := Temp.GetIDsOfNames(@Name, 1, @MemberID);
          if Code <> DISP_E_UNKNOWNNAME then
          begin
            OleCheck(Code);
            Exchange(TypeInfo, Temp);
            Result := True;
            Break;
          end;
        end;
        Inc(I);
      end;
    finally
      SysFreeString(Name);
    end;
  end;

begin
  HelpCtx := 0;
  HelpFile := '';
  CreateControl;
  OleCheck((FOleObject as IProvideClassInfo).GetClassInfo(TypeInfo));
  MemberID := MEMBERID_NIL;
  if Length(Member) > 0 then
  begin
    OleCheck(TypeInfo.GetTypeAttr(TypeAttr));
    ImplTypes := TypeAttr.cImplTypes;
    TypeInfo.ReleaseTypeAttr(TypeAttr);
    Result := Find(Member, TypeInfo);
    if (not Result) and (Member[Length(Member)] = '_') then
    begin
      Delete(Member, Length(Member)-1, 1);
      Result := Find(Member, TypeInfo);
    end;
    if (not Result) and (Pos('On', Member) = 1) then
    begin
      Delete(Member, 1, 2);
      Result := Find(Member, TypeInfo);
    end;
    if not Result then Exit;
  end;
  OleCheck(TypeInfo.GetDocumentation(MemberID, nil, nil, @HelpCtx, @HlpFile));
  HelpFile := OleStrToString(HlpFile);
  SysFreeString(HlpFile);
  Result := True;
end;

function TpsvApplet.GetIDispatchProp(Index: Integer): IDispatch;
var
  Temp: TVarData;
begin
  GetProperty(Index, Temp);
  Result := IDispatch(Temp.VDispatch);
end;

function TpsvApplet.GetIntegerProp(Index: Integer): Integer;
var
  Temp: TVarData;
begin
  GetProperty(Index, Temp);
  Result := Temp.VInteger;
end;

function TpsvApplet.GetIUnknownProp(Index: Integer): IUnknown;
var
  Temp: TVarData;
begin
  GetProperty(Index, Temp);
  Result := IUnknown(Temp.VUnknown);
end;

function TpsvApplet.GetMainMenu: TMainMenu;
var
  Form: TCustomForm;
begin
  Result := nil;
  Form := GetParentForm(Self);
  if Form <> nil then
    if (Form is TForm) and (TForm(Form).FormStyle <> fsMDIChild) then
      Result := Form.Menu
    else
      if Application.MainForm <> nil then
        Result := Application.MainForm.Menu;
end;

procedure TpsvApplet.GetObjectVerbs(List: TStrings);
var
  EnumOleVerb: IEnumOleVerb;
  OleVerb: TOleVerb;
  Code: HResult;
begin
  CreateControl;
  List.Clear;
  Code := FOleObject.EnumVerbs(EnumOleVerb);
  if Code = OLE_S_USEREG then
    Code := OleRegEnumVerbs(FControlData.ClassID, EnumOleVerb);
  if Code = 0 then
    while (EnumOleVerb.Next(1, OleVerb, nil) = 0) do
      if (OleVerb.grfAttribs and OLEVERBATTRIB_ONCONTAINERMENU <> 0) then
      begin
        List.AddObject(StripHotkey(OleVerb.lpszVerbName), TObject(OleVerb.lVerb));
      end;
end;

function TpsvApplet.GetWordBoolProp(Index: Integer): WordBool;
var
  Temp: TVarData;
begin
  GetProperty(Index, Temp);
  Result := Temp.VBoolean;
end;

function TpsvApplet.GetTDateTimeProp(Index: Integer): TDateTime;
var
  Temp: TVarData;
begin
  GetProperty(Index, Temp);
  Result := Temp.VDate;
end;

function TpsvApplet.GetTFontProp(Index: Integer): TFont;
var
  I: Integer;
begin
  Result := nil;
  for I := 0 to FFonts.Count-1 do
    if FControlData^.FontIDs^[I] = Index then
    begin
      Result := TFont(FFonts[I]);
      if Result.FontAdapter = nil then
        SetOleFont(Result, GetIDispatchProp(Index) as IFontDisp);
    end;
end;

function TpsvApplet.GetOleBoolProp(Index: Integer): TOleBool;
var
  Temp: TVarData;
begin
  GetProperty(Index, Temp);
  Result := Temp.VBoolean;
end;

function TpsvApplet.GetOleDateProp(Index: Integer): TOleDate;
var
  Temp: TVarData;
begin
  GetProperty(Index, Temp);
  Result := Temp.VDate;
end;

function TpsvApplet.GetOleEnumProp(Index: Integer): TOleEnum;
begin
  Result := GetIntegerProp(Index);
end;

function TpsvApplet.GetTOleEnumProp(Index: Integer): TOleEnum;
begin
  Result := GetIntegerProp(Index);
end;


function TpsvApplet.GetDefaultDispatch: IDispatch;
begin
  CreateControl;
  Result := FOleObject as IDispatch;
end;

function TpsvApplet.GetOleVariantProp(Index: Integer): OleVariant;
begin
  VarClear(Result);
  GetProperty(Index, TVarData(Result));
end;

function TpsvApplet.GetTPictureProp(Index: Integer): TPicture;
var
  I: Integer;
begin
  Result := nil;
  for I := 0 to FPictures.Count-1 do
    if FControlData^.PictureIDs^[I] = Index then
    begin
      Result := TPicture(FPictures[I]);
      if Result.PictureAdapter = nil then
        SetOlePicture(Result, GetIDispatchProp(Index) as IPictureDisp);
    end;
end;


function TpsvApplet.GetPropDisplayString(DispID: Integer): string;
var
  S: WideString;
begin
  CreateControl;
  if (FPropBrowsing <> nil) and
    (FPropBrowsing.GetDisplayString(DispID, S) = 0) then
    Result := S else
    Result := GetStringProp(DispID);
end;

procedure TpsvApplet.GetPropDisplayStrings(DispID: Integer; List: TStrings);
var
  Strings: TCAPOleStr;
  Cookies: TCALongint;
  I: Integer;
begin
  CreateControl;
  List.Clear;
  if (FPropBrowsing <> nil) and
    (FPropBrowsing.GetPredefinedStrings(DispID, Strings, Cookies) = 0) then
    try
      for I := 0 to Strings.cElems - 1 do
        List.AddObject(Strings.pElems^[I], TObject(Cookies.pElems^[I]));
    finally
      for I := 0 to Strings.cElems - 1 do
        CoTaskMemFree(Strings.pElems^[I]);
      CoTaskMemFree(Strings.pElems);
      CoTaskMemFree(Cookies.pElems);
    end;
end;

var  // init to zero, never written to
  DispParams: TDispParams = ();

procedure TpsvApplet.GetProperty(Index: Integer; var Value: TVarData);
var
  Status: HResult;
  ExcepInfo: TExcepInfo;
begin
  CreateControl;
  Value.VType := varEmpty;
  Status := FControlDispatch.Invoke(Index, GUID_NULL, 0,
    DISPATCH_PROPERTYGET, DispParams, @Value, @ExcepInfo, nil);
  if Status <> 0 then DispatchInvokeError(Status, ExcepInfo);
end;

function TpsvApplet.GetShortIntProp(Index: Integer): ShortInt;
begin
  Result := GetIntegerProp(Index);
end;

function TpsvApplet.GetSingleProp(Index: Integer): Single;
var
  Temp: TVarData;
begin
  GetProperty(Index, Temp);
  Result := Temp.VSingle;
end;

function TpsvApplet.GetSmallintProp(Index: Integer): Smallint;
var
  Temp: TVarData;
begin
  GetProperty(Index, Temp);
  Result := Temp.VSmallint;
end;

function TpsvApplet.GetStringProp(Index: Integer): string;
begin
  Result := GetVariantProp(Index);
end;

function TpsvApplet.GetVariantProp(Index: Integer): Variant;
begin
  Result := GetOleVariantProp(Index);
end;

function TpsvApplet.GetWideStringProp(Index: Integer): WideString;
var
  Temp: TVarData;
begin
  Result := '';
  GetProperty(Index, Temp);
  Pointer(Result) := Temp.VOleStr;
end;

function TpsvApplet.GetWordProp(Index: Integer): Word;
begin
  Result := GetIntegerProp(Index);
end;

procedure TpsvApplet.HookControlWndProc;
var
  WndHandle: HWnd;
begin
  if (FOleInPlaceObject <> nil) and (WindowHandle = 0) then
  begin
    WndHandle := 0;
    FOleInPlaceObject.GetWindow(WndHandle);
    if WndHandle = 0 then raise EOleError.CreateRes(@SNoWindowHandle);
    WindowHandle := WndHandle;
    DefWndProc := Pointer(GetWindowLong(WindowHandle, GWL_WNDPROC));
    CreationControl := Self;
    SetWindowLong(WindowHandle, GWL_WNDPROC, Longint(@InitWndProc));
    SendMessage(WindowHandle, WM_NULL, 0, 0);
  end;
end;


type
  PVarArg = ^TVarArg;
  TVarArg = array[0..3] of DWORD;



procedure GetStringResult(BStr: TBStr; var Result: string);
begin
  try
    OleStrToStrVar(BStr, Result);
  finally
    SysFreeString(BStr);
  end;
end;


procedure TpsvApplet.InvokeMethod(const DispInfo; Result: Pointer); assembler;
asm
        PUSH    EBX
        PUSH    ESI
        PUSH    EDI
        MOV     EBX,EAX
        MOV     ESI,EDX
        MOV     EDI,ECX
        CALL    TpsvApplet.CreateControl
        PUSH    [ESI].TDispInfo.DispID
        MOV     ECX,ESP
        XOR     EAX,EAX
        PUSH    EAX
        PUSH    EAX
        PUSH    EAX
        PUSH    EAX
        MOV     EDX,ESP
        LEA     EAX,[EBP+16]
        CMP     [ESI].TDispInfo.ResType,varOleStr
        JE      @@1
        CMP     [ESI].TDispInfo.ResType,varVariant
        JE      @@1
        LEA     EAX,[EBP+12]
@@1:    PUSH    EAX
        PUSH    EDX
        LEA     EDX,[ESI].TDispInfo.CallDesc
        MOV     EAX,[EBX].TpsvApplet.FControlDispatch
        CALL    DispatchInvoke
        XOR     EAX,EAX
        MOV     AL,[ESI].TDispInfo.ResType
        JMP     @ResultTable.Pointer[EAX*4]

@ResultTable:
        DD      @ResEmpty
        DD      @ResNull
        DD      @ResSmallint
        DD      @ResInteger
        DD      @ResSingle
        DD      @ResDouble
        DD      @ResCurrency
        DD      @ResDate
        DD      @ResString
        DD      @ResDispatch
        DD      @ResError
        DD      @ResBoolean
        DD      @ResVariant

@ResSmallint:
@ResBoolean:
        MOV     AX,[ESP+8]
        MOV     [EDI],AX
        JMP     @ResDone

@ResString:
        MOV     EAX,[ESP+8]
        MOV     EDX,EDI
        CALL    GetStringResult
        JMP     @ResDone

@ResVariant:
        MOV     EAX,EDI
        CALL    System.@VarClear
        MOV     ESI,ESP
        MOV     ECX,4
        REP     MOVSD
        JMP     @ResDone

@ResDouble:
@ResCurrency:
@ResDate:
        MOV     EAX,[ESP+12]
        MOV     [EDI+4],EAX

@ResInteger:
@ResSingle:
        MOV     EAX,[ESP+8]
        MOV     [EDI],EAX

@ResEmpty:
@ResNull:
@ResDispatch:
@ResError:
@ResDone:
        ADD     ESP,20
        POP     EDI
        POP     ESI
        POP     EBX
end;

function TpsvApplet.IsCustomProperty(DispID: Integer): Boolean;
var
  W: WideString;
begin
  Result := (FPropBrowsing <> nil) and
    (FPropBrowsing.GetDisplayString(DispID, W) = 0);
end;

function TpsvApplet.IsPropPageProperty(DispID: Integer): Boolean;
var
  PPID: TCLSID;
begin
  Result := (FPropBrowsing <> nil) and
    (FPropBrowsing.MapPropertyToPage(DispID, PPID) = S_FALSE) and not
    IsEqualCLSID(PPID, GUID_NULL);
end;

function TpsvApplet.PaletteChanged(Foreground: Boolean): Boolean;
begin
  Result := False;
  if HandleAllocated and Foreground then
    Result := CallWindowProc(DefWndProc, Handle, WM_QUERYNEWPALETTE, 0, 0) <> 0;
  if not Result then
    Result := inherited PaletteChanged(Foreground);
end;

procedure TpsvApplet.PictureChanged(Sender: TObject);
var
  I: Integer;
begin
  if (FPictures = nil) or not (Sender is TPicture) then Exit;
  for I := 0 to FPictures.Count - 1 do
    if FPictures[I] = Sender then
    begin
      if (TPicture(Sender).PictureAdapter <> nil) then
        SetTPictureProp(FControlData.PictureIDs^[I], TPicture(Sender));
      Exit;
    end;
end;


procedure TpsvApplet.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  if ((AWidth <> Width) and (Width > 0)) or ((AHeight <> Height) and (Height > 0)) then
    if (FMiscStatus and OLEMISC_INVISIBLEATRUNTIME <> 0) or
      ((FOleObject.SetExtent(DVASPECT_CONTENT, Point(
        MulDiv(AWidth, 2540, Screen.PixelsPerInch),
        MulDiv(AHeight, 2540, Screen.PixelsPerInch))) <> S_OK)) then
    begin
      AWidth := Width;
      AHeight := Height;
    end;
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
end;

procedure TpsvApplet.SetByteProp(Index: Integer; Value: Byte);
begin
  SetIntegerProp(Index, Value);
end;

procedure TpsvApplet.SetColorProp(Index: Integer; Value: TColor);
begin
  SetIntegerProp(Index, Value);
end;

procedure TpsvApplet.SetTColorProp(Index: Integer; Value: TColor);
begin
  SetIntegerProp(Index, Value);
end;

procedure TpsvApplet.SetCompProp(Index: Integer; const Value: Comp);
var
  Temp: TVarData;
begin
  Temp.VType := VT_I8;
  Temp.VDouble := Value;
  SetProperty(Index, Temp);
end;

procedure TpsvApplet.SetCurrencyProp(Index: Integer; const Value: Currency);
var
  Temp: TVarData;
begin
  Temp.VType := varCurrency;
  Temp.VCurrency := Value;
  SetProperty(Index, Temp);
end;

procedure TpsvApplet.SetDoubleProp(Index: Integer; const Value: Double);
var
  Temp: TVarData;
begin
  Temp.VType := varDouble;
  Temp.VDouble := Value;
  SetProperty(Index, Temp);
end;

procedure TpsvApplet.SetIDispatchProp(Index: Integer; const Value: IDispatch);
var
  Temp: TVarData;
begin
  Temp.VType := varDispatch;
  Temp.VDispatch := Pointer(Value);
  SetProperty(Index, Temp);
end;

procedure TpsvApplet.SetIntegerProp(Index: Integer; Value: Integer);
var
  Temp: TVarData;
begin
  Temp.VType := varInteger;
  Temp.VInteger := Value;
  SetProperty(Index, Temp);
end;

procedure TpsvApplet.SetIUnknownProp(Index: Integer; const Value: IUnknown);
var
  Temp: TVarData;
begin
  Temp.VType := VT_UNKNOWN;
  Temp.VUnknown := Pointer(Value);
  SetProperty(Index, Temp);
end;

procedure TpsvApplet.SetName(const Value: TComponentName);
var
  OldName: string;
  DispID: Integer;
begin
  OldName := Name;
  inherited SetName(Value);
  if FOleControl <> nil then
  begin
    FOleControl.OnAmbientPropertyChange(DISPID_AMBIENT_DISPLAYNAME);
    if FControlData^.Flags and (cfCaption or cfText) <> 0 then
    begin
      if FControlData^.Flags and cfCaption <> 0 then
        DispID := DISPID_CAPTION else
        DispID := DISPID_TEXT;
      if OldName = GetStringProp(DispID) then SetStringProp(DispID, Value);
    end;
  end;
end;

procedure TpsvApplet.SetWordBoolProp(Index: Integer; Value: WordBool);
var
  Temp: TVarData;
begin
  Temp.VType := varBoolean;
  if Value then
    Temp.VBoolean := WordBool(-1) else
    Temp.VBoolean := WordBool(0);
  SetProperty(Index, Temp);
end;

procedure TpsvApplet.SetTDateTimeProp(Index: Integer; const Value: TDateTime);
var
  Temp: TVarData;
begin
  Temp.VType := varDate;
  Temp.VDate := Value;
  SetProperty(Index, Temp);
end;

procedure TpsvApplet.SetTFontProp(Index: Integer; Value: TFont);
var
  I: Integer;
  F: TFont;
  Temp: IFontDisp;
begin
  for I := 0 to FFonts.Count-1 do
    if FControlData^.FontIDs^[I] = Index then
    begin
      F := TFont(FFonts[I]);
      F.Assign(Value);
      if F.FontAdapter = nil then
      begin
        GetOleFont(F, Temp);
        SetIDispatchProp(Index, Temp);
      end;
    end;
end;

procedure TpsvApplet.SetOleBoolProp(Index: Integer; Value: TOleBool);
var
  Temp: TVarData;
begin
  Temp.VType := varBoolean;
  if Value then
    Temp.VBoolean := WordBool(-1) else
    Temp.VBoolean := WordBool(0);
  SetProperty(Index, Temp);
end;

procedure TpsvApplet.SetOleDateProp(Index: Integer; const Value: TOleDate);
var
  Temp: TVarData;
begin
  Temp.VType := varDate;
  Temp.VDate := Value;
  SetProperty(Index, Temp);
end;

procedure TpsvApplet.SetOleEnumProp(Index: Integer; Value: TOleEnum);
begin
  SetIntegerProp(Index, Value);
end;

procedure TpsvApplet.SetTOleEnumProp(Index: Integer; Value: TOleEnum);
begin
  SetIntegerProp(Index, Value);
end;

procedure TpsvApplet.SetOleVariantProp(Index: Integer; const Value: OleVariant);
begin
  SetProperty(Index, TVarData(Value));
end;

procedure TpsvApplet.SetParent(AParent: TWinControl);
var
  CS: IOleClientSite;
  X: Integer;
begin
  inherited SetParent(AParent);
  if (AParent <> nil) then
  begin
    try  // work around ATL bug
      X := FOleObject.GetClientSite(CS);
    except
      X := -1;
    end;
    if (X <> 0) or (CS = nil) then
      OleCheck(FOleObject.SetClientSite(Self));
    if FOleControl <> nil then
      FOleControl.OnAmbientPropertyChange(DISPID_UNKNOWN);
  end;
end;

procedure TpsvApplet.SetTPictureProp(Index: Integer; Value: TPicture);
var
  I: Integer;
  P: TPicture;
  Temp: IPictureDisp;
begin
  if FUpdatingPictures then Exit;
  FUpdatingPictures := True;
  try
    for I := 0 to FPictures.Count-1 do
      if FControlData^.PictureIDs^[I] = Index then
      begin
        P := TPicture(FPictures[I]);
        P.Assign(Value);
        GetOlePicture(P, Temp);
        SetIDispatchProp(Index, Temp);
      end;
  finally
    FUpdatingPictures := False;
  end;
end;

procedure TpsvApplet.SetPropDisplayString(DispID: Integer;
  const Value: string);
var
  I: Integer;
  Values: TStringList;
  V: OleVariant;
begin
  Values := TStringList.Create;
  try
    GetPropDisplayStrings(DispID, Values);
    for I := 0 to Values.Count - 1 do
      if AnsiCompareText(Value, Values[I]) = 0 then
      begin
        OleCheck(FPropBrowsing.GetPredefinedValue(DispID,
          Integer(Values.Objects[I]), V));
        SetProperty(DispID, TVarData(V));
        Exit;
      end;
  finally
    Values.Free;
  end;
  SetStringProp(DispID, Value);
end;

procedure TpsvApplet.SetProperty(Index: Integer; const Value: TVarData);
const
  DispIDArgs: Longint = DISPID_PROPERTYPUT;
var
  Status, InvKind: Integer;
  DispParams: TDispParams;
  ExcepInfo: TExcepInfo;
begin
  CreateControl;
  DispParams.rgvarg := @Value;
  DispParams.rgdispidNamedArgs := @DispIDArgs;
  DispParams.cArgs := 1;
  DispParams.cNamedArgs := 1;
  if Value.VType <> varDispatch then
    InvKind := DISPATCH_PROPERTYPUT else
    InvKind := DISPATCH_PROPERTYPUTREF;
  Status := FControlDispatch.Invoke(Index, GUID_NULL, 0,
    InvKind, DispParams, nil, @ExcepInfo, nil);
  if Status <> 0 then DispatchInvokeError(Status, ExcepInfo);
end;

procedure TpsvApplet.SetShortintProp(Index: Integer; Value: ShortInt);
begin
  SetIntegerProp(Index, Value);
end;

procedure TpsvApplet.SetSingleProp(Index: Integer; const Value: Single);
var
  Temp: TVarData;
begin
  Temp.VType := varSingle;
  Temp.VSingle := Value;
  SetProperty(Index, Temp);
end;

procedure TpsvApplet.SetSmallintProp(Index: Integer; Value: Smallint);
var
  Temp: TVarData;
begin
  Temp.VType := varSmallint;
  Temp.VSmallint := Value;
  SetProperty(Index, Temp);
end;

procedure TpsvApplet.SetStringProp(Index: Integer; const Value: string);
var
  Temp: TVarData;
begin
  Temp.VType := varOleStr;
  Temp.VOleStr := StringToOleStr(Value);
  try
    SetProperty(Index, Temp);
  finally
    SysFreeString(Temp.VOleStr);
  end;
end;

procedure TpsvApplet.SetUIActive(Active: Boolean);
var
  Form: TCustomForm;
begin
  Form := GetParentForm(Self);
  if Form <> nil then
    if Active then
    begin
      if (Form.ActiveOleControl <> nil) and
        (Form.ActiveOleControl <> Self) then
        Form.ActiveOleControl.Perform(CM_UIDEACTIVATE, 0, 0);
      Form.ActiveOleControl := Self;
    end else
      if Form.ActiveOleControl = Self then Form.ActiveOleControl := nil;
end;

procedure TpsvApplet.SetVariantProp(Index: Integer; const Value: Variant);
begin
  SetOleVariantProp(Index, Value);
end;

procedure TpsvApplet.SetWideStringProp(Index: Integer; const Value: WideString);
var
  Temp: TVarData;
begin
  Temp.VType := varOleStr;
  if Value <> '' then
    Temp.VOleStr := PWideChar(Value)
  else
    Temp.VOleStr := nil;
  SetProperty(Index, Temp);
end;

procedure TpsvApplet.SetWordProp(Index: Integer; Value: Word);
begin
  SetIntegerProp(Index, Value);
end;


procedure TpsvApplet.WndProc(var Message: TMessage);
var
  WinMsg: TMsg;
begin
  if (Message.Msg >= CN_BASE + WM_KEYFIRST) and
    (Message.Msg <= CN_BASE + WM_KEYLAST) and
    (FOleInPlaceActiveObject <> nil) then
  begin
    WinMsg.HWnd := Handle;
    WinMsg.Message := Message.Msg - CN_BASE;
    WinMsg.WParam := Message.WParam;
    WinMsg.LParam := Message.LParam;
    WinMsg.Time := GetMessageTime;
    WinMsg.Pt.X := $115DE1F1;
    WinMsg.Pt.Y := $115DE1F1;
    if FOleInPlaceActiveObject.TranslateAccelerator(WinMsg) = S_OK then
    begin
      Message.Result := 1;
      Exit;
    end;
  end;
  case TMessage(Message).Msg of
    CM_PARENTFONTCHANGED:
      if ParentFont and (FOleControl <> nil) then
      begin
        FOleControl.OnAmbientPropertyChange(DISPID_AMBIENT_FONT);
        FOleControl.OnAmbientPropertyChange(DISPID_AMBIENT_FORECOLOR);
      end;
    CM_PARENTCOLORCHANGED:
      if ParentColor and (FOleControl <> nil) then
        FOleControl.OnAmbientPropertyChange(DISPID_AMBIENT_BACKCOLOR);
  end;
  inherited WndProc(Message);
end;


procedure TpsvApplet.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  if FMiscStatus and OLEMISC_INVISIBLEATRUNTIME = 0 then
    DefaultHandler(Message) else
    inherited;
end;

procedure TpsvApplet.WMPaint(var Message: TWMPaint);
var
  DC: HDC;
  PS: TPaintStruct;
begin
  if FMiscStatus and OLEMISC_INVISIBLEATRUNTIME <> 0 then
  begin
    DC := Message.DC;
    if DC = 0 then DC := BeginPaint(Handle, PS);
    OleDraw(FOleObject, DVASPECT_CONTENT, DC, ClientRect);
    if Message.DC = 0 then EndPaint(Handle, PS);
  end else
    inherited;
end;

procedure TpsvApplet.CMDocWindowActivate(var Message: TMessage);
var
  Form: TCustomForm;
  F: TForm;
begin
  Form := GetParentForm(Self);
  F := nil;
  if Form is TForm then F := TForm(Form);
  if (F <> nil) {and (F.FormStyle = fsMDIChild)} then
  begin
    FOleInPlaceActiveObject.OnDocWindowActivate(LongBool(Message.WParam));
    if Message.WParam = 0 then SetMenu(0, 0, 0);
  end;
end;

procedure TpsvApplet.CMColorChanged(var Message: TMessage);
begin
  inherited;
  if (FControlData^.Flags and cfBackColor <> 0) and not FUpdatingColor and
    HandleAllocated then
  begin
    FUpdatingColor := True;
    try
      SetColorProp(DISPID_BACKCOLOR, Color);
    finally
      FUpdatingColor := False;
    end;
  end;
end;

procedure TpsvApplet.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if (FControlData^.Flags and cfEnabled <> 0) and not FUpdatingEnabled and
    HandleAllocated then
  begin
    FUpdatingEnabled := True;
    try
      SetWordBoolProp(DISPID_ENABLED, Enabled);
    finally
      FUpdatingEnabled := False;
    end;
  end;
end;

procedure TpsvApplet.CMFontChanged(var Message: TMessage);
begin
  inherited;
  if (FControlData^.Flags and (cfForeColor or cfFont) <> 0) and
    not FUpdatingFont and HandleAllocated then
  begin
    FUpdatingFont := True;
    try
      if FControlData^.Flags and cfForeColor <> 0 then
        SetIntegerProp(DISPID_FORECOLOR, Font.Color);
      if FControlData^.Flags and cfFont <> 0 then
        SetVariantProp(DISPID_FONT, FontToOleFont(Font));
    finally
      FUpdatingFont := False;
    end;
  end;
end;

procedure TpsvApplet.CMDialogKey(var Message: TMessage);
var
  Info: TControlInfo;
  Msg: TMsg;
  Cmd: Word;
begin
  if CanFocus then
  begin
    Info.cb := SizeOf(Info);
    if (FOleControl.GetControlInfo(Info) = S_OK) and (Info.cAccel <> 0) then
    begin
      FillChar(Msg, SizeOf(Msg), 0);
      Msg.hwnd := Handle;
      Msg.message := WM_KEYDOWN;
      Msg.wParam := Message.WParam;
      Msg.lParam := Message.LParam;
      if IsAccelerator(Info.hAccel, Info.cAccel, @Msg, Cmd) then
      begin
        FOleControl.OnMnemonic(@Msg);
        Message.Result := 1;
        Exit;
      end;
    end;
  end;
  inherited;
end;

procedure TpsvApplet.CMUIActivate(var Message: TMessage);
var
  F: TCustomForm;
begin
  F := GetParentForm(Self);
  if (F = nil) or (F.ActiveOleControl <> Self) then
    FOleObject.DoVerb(OLEIVERB_UIACTIVATE, nil, Self, 0,
      GetParentHandle, BoundsRect);
end;

procedure TpsvApplet.CMUIDeactivate(var Message: TMessage);
var
  F: TCustomForm;
begin
  F := GetParentForm(Self);
  if (F = nil) or (F.ActiveOleControl = Self) then
  begin
    if FOleInPlaceObject <> nil then FOleInPlaceObject.UIDeactivate;
    if (F <> nil) and (F.ActiveControl = Self) then OnUIDeactivate(False);
  end;
end;

{ TpsvApplet.IUnknown }

function TpsvApplet.QueryInterface(const IID: TGUID; out Obj): HResult;
begin
  if GetInterface(IID, Obj) then Result := S_OK else Result := E_NOINTERFACE;
end;

function TpsvApplet._AddRef: Integer;
begin
  Inc(FRefCount);
  Result := FRefCount;
end;

function TpsvApplet._Release: Integer;
begin
  Dec(FRefCount);
  Result := FRefCount;
end;

{ TpsvApplet.IOleClientSite }

function TpsvApplet.SaveObject: HResult;
begin
  Result := S_OK;
end;

function TpsvApplet.GetMoniker(dwAssign: Longint; dwWhichMoniker: Longint;
  out mk: IMoniker): HResult;
begin
  Result := E_NOTIMPL;
end;

function TpsvApplet.GetContainer(out container: IOleContainer): HResult;
begin
  Result := E_NOINTERFACE;
end;

function TpsvApplet.ShowObject: HResult;
begin
  HookControlWndProc;
  Result := S_OK;
end;

function TpsvApplet.OnShowWindow(fShow: BOOL): HResult;
begin
  Result := S_OK;
end;

function TpsvApplet.RequestNewObjectLayout: HResult;
var
  Extent: TPoint;
  W, H: Integer;
begin
  Result := FOleObject.GetExtent(DVASPECT_CONTENT, Extent);
  if Result <> S_OK then Exit;
  W := MulDiv(Extent.X, Screen.PixelsPerInch, 2540);
  H := MulDiv(Extent.Y, Screen.PixelsPerInch, 2540);
  if (FMiscStatus and OLEMISC_INVISIBLEATRUNTIME <> 0) and (FOleControl = nil) then
  begin
    if W > 32 then W := 32;
    if H > 32 then H := 32;
  end;
  SetBounds(Left, Top, W, H);
end;

{ TpsvApplet.IOleControlSite }

function TpsvApplet.OnControlInfoChanged: HResult;
begin
  Result := E_NOTIMPL;
end;

function TpsvApplet.LockInPlaceActive(fLock: BOOL): HResult;
begin
  Result := E_NOTIMPL;
end;

function TpsvApplet.GetExtendedControl(out disp: IDispatch): HResult;
begin
  Result := E_NOTIMPL;
end;

function TpsvApplet.TransformCoords(var ptlHimetric: TPoint;
  var ptfContainer: TPointF; flags: Longint): HResult;
begin
  if flags and XFORMCOORDS_HIMETRICTOCONTAINER <> 0 then
  begin
    ptfContainer.X := MulDiv(ptlHimetric.X, Screen.PixelsPerInch, 2540);
    ptfContainer.Y := MulDiv(ptlHimetric.Y, Screen.PixelsPerInch, 2540);
  end else
  begin
    ptlHimetric.X := Integer(Round(ptfContainer.X * 2540 / Screen.PixelsPerInch));
    ptlHimetric.Y := Integer(Round(ptfContainer.Y * 2540 / Screen.PixelsPerInch));
  end;
  Result := S_OK;
end;

function TpsvApplet.OleControlSite_TranslateAccelerator(
  msg: PMsg; grfModifiers: Longint): HResult;
begin
  Result := E_NOTIMPL;
end;

function TpsvApplet.OnFocus(fGotFocus: BOOL): HResult;
begin
  Result := E_NOTIMPL;
end;

function TpsvApplet.ShowPropertyFrame: HResult;
begin
  Result := E_NOTIMPL;
end;

{ TpsvApplet.IOleWindow }

function TpsvApplet.ContextSensitiveHelp(fEnterMode: BOOL): HResult;
begin
  Result := S_OK;
end;

{ TpsvApplet.IOleInPlaceSite }

function TpsvApplet.OleInPlaceSite_GetWindow(out wnd: HWnd): HResult;
begin
  Result := S_OK;
  wnd := GetParentHandle;
  if wnd = 0 then Result := E_FAIL;
end;

function TpsvApplet.CanInPlaceActivate: HResult;
begin
  Result := S_OK;
end;

function TpsvApplet.OnInPlaceActivate: HResult;
begin
  FOleObject.QueryInterface(IOleInPlaceObject, FOleInPlaceObject);
  FOleObject.QueryInterface(IOleInPlaceActiveObject, FOleInPlaceActiveObject);
  Result := S_OK;
end;

function TpsvApplet.OnUIActivate: HResult;
begin
  SetUIActive(True);
  Result := S_OK;
end;

function TpsvApplet.GetWindowContext(out frame: IOleInPlaceFrame;
  out doc: IOleInPlaceUIWindow; out rcPosRect: TRect;
  out rcClipRect: TRect; out frameInfo: TOleInPlaceFrameInfo): HResult;
begin
  frame := Self;
  doc := nil;
  rcPosRect := BoundsRect;
  SetRect(rcClipRect, 0, 0, 32767, 32767);
  with frameInfo do
  begin
    fMDIApp := False;
    hWndFrame := GetTopParentHandle;
    hAccel := 0;
    cAccelEntries := 0;
  end;
  Result := S_OK;
end;

function TpsvApplet.Scroll(scrollExtent: TPoint): HResult;
begin
  Result := E_NOTIMPL;
end;

function TpsvApplet.OnUIDeactivate(fUndoable: BOOL): HResult;
begin
  SetMenu(0, 0, 0);
  SetUIActive(False);
  Result := S_OK;
end;

function TpsvApplet.OnInPlaceDeactivate: HResult;
begin
  FOleInPlaceActiveObject := nil;
  FOleInPlaceObject := nil;
  Result := S_OK;
end;

function TpsvApplet.DiscardUndoState: HResult;
begin
  Result := E_NOTIMPL;
end;

function TpsvApplet.DeactivateAndUndo: HResult;
begin
  FOleInPlaceObject.UIDeactivate;
  Result := S_OK;
end;

function TpsvApplet.OnPosRectChange(const rcPosRect: TRect): HResult;
begin
  FOleInPlaceObject.SetObjectRects(rcPosRect, Rect(0, 0, 32767, 32767));
  Result := S_OK;
end;

{ TpsvApplet.IOleInPlaceUIWindow }

function TpsvApplet.GetBorder(out rectBorder: TRect): HResult;
begin
  Result := INPLACE_E_NOTOOLSPACE;
end;

function TpsvApplet.RequestBorderSpace(const borderwidths: TRect): HResult;
begin
  Result := INPLACE_E_NOTOOLSPACE;
end;

function TpsvApplet.SetBorderSpace(pborderwidths: PRect): HResult;
begin
  Result := E_NOTIMPL;
end;

function TpsvApplet.SetActiveObject(const activeObject: IOleInPlaceActiveObject;
  pszObjName: POleStr): HResult;
begin
  Result := S_OK;
end;

{ TpsvApplet.IOleInPlaceFrame }

function TpsvApplet.OleInPlaceFrame_GetWindow(out wnd: HWnd): HResult;
begin
  wnd := GetTopParentHandle;
  Result := S_OK;
end;

function TpsvApplet.InsertMenus(hmenuShared: HMenu;
  var menuWidths: TOleMenuGroupWidths): HResult;
var
  Menu: TMainMenu;
begin
  Menu := GetMainMenu;
  if Menu <> nil then
    Menu.PopulateOle2Menu(hmenuShared, [0, 2, 4], menuWidths.width);
  Result := S_OK;
end;

function TpsvApplet.SetMenu(hmenuShared: HMenu; holemenu: HMenu;
  hwndActiveObject: HWnd): HResult;
var
  Menu: TMainMenu;
begin
  Menu := GetMainMenu;
  Result := S_OK;
  if Menu <> nil then
  begin
    Menu.SetOle2MenuHandle(hmenuShared);
    Result := OleSetMenuDescriptor(holemenu, Menu.WindowHandle,
      hwndActiveObject, nil, nil);
  end;
end;

function TpsvApplet.RemoveMenus(hmenuShared: HMenu): HResult;
begin
  while GetMenuItemCount(hmenuShared) > 0 do
    RemoveMenu(hmenuShared, 0, MF_BYPOSITION);
  Result := S_OK;
end;

function TpsvApplet.SetStatusText(pszStatusText: POleStr): HResult;
begin
  Result := S_OK;
end;

function TpsvApplet.EnableModeless(fEnable: BOOL): HResult;
begin
  Result := S_OK;
end;

function TpsvApplet.OleInPlaceFrame_TranslateAccelerator(
  var msg: TMsg; wID: Word): HResult;
begin
  Result := S_FALSE;
end;

{ TpsvApplet.IDispatch }

function TpsvApplet.GetTypeInfoCount(out Count: Integer): HResult;
begin
  Count := 0;
  Result := S_OK;
end;

function TpsvApplet.GetTypeInfo(Index, LocaleID: Integer;
  out TypeInfo): HResult;
begin
  Pointer(TypeInfo) := nil;
  Result := E_NOTIMPL;
end;

function TpsvApplet.GetIDsOfNames(const IID: TGUID; Names: Pointer;
  NameCount, LocaleID: Integer; DispIDs: Pointer): HResult;
begin
  Result := E_NOTIMPL;
end;

function TpsvApplet.Invoke(DispID: Integer; const IID: TGUID;
  LocaleID: Integer; Flags: Word; var Params;
  VarResult, ExcepInfo, ArgErr: Pointer): HResult;
var
  F: TFont;
begin
  if (Flags and DISPATCH_PROPERTYGET <> 0) and (VarResult <> nil) then
  begin
    Result := S_OK;
    case DispID of
      DISPID_AMBIENT_BACKCOLOR:
        PVariant(VarResult)^ := Color;
      DISPID_AMBIENT_DISPLAYNAME:
        PVariant(VarResult)^ := StringToVarOleStr(Name);
      DISPID_AMBIENT_FONT:
      begin
        if (Parent <> nil) and ParentFont then
          F := TpsvApplet(Parent).Font
        else
          F := Font;
        PVariant(VarResult)^ := FontToOleFont(F);
      end;
      DISPID_AMBIENT_FORECOLOR:
        PVariant(VarResult)^ := Font.Color;
      DISPID_AMBIENT_LOCALEID:
        PVariant(VarResult)^ := Integer(GetUserDefaultLCID);
      DISPID_AMBIENT_MESSAGEREFLECT:
        PVariant(VarResult)^ := True;
      DISPID_AMBIENT_USERMODE:
        PVariant(VarResult)^ := not (csDesigning in ComponentState);
      DISPID_AMBIENT_UIDEAD:
        PVariant(VarResult)^ := csDesigning in ComponentState;
      DISPID_AMBIENT_SHOWGRABHANDLES:
        PVariant(VarResult)^ := False;
      DISPID_AMBIENT_SHOWHATCHING:
        PVariant(VarResult)^ := False;
      DISPID_AMBIENT_SUPPORTSMNEMONICS:
        PVariant(VarResult)^ := True;
      DISPID_AMBIENT_AUTOCLIP:
        PVariant(VarResult)^ := True;
    else
      Result := DISP_E_MEMBERNOTFOUND;
    end;
  end else
    Result := DISP_E_MEMBERNOTFOUND;
end;

{ TpsvApplet.IPropertyNotifySink }

function TpsvApplet.OnChanged(dispid: TDispID): HResult;
begin
  try
    case dispid of
      DISPID_BACKCOLOR:
        if not FUpdatingColor then
        begin
          FUpdatingColor := True;
          try
            Color := GetIntegerProp(DISPID_BACKCOLOR);
          finally
            FUpdatingColor := False;
          end;
        end;
      DISPID_ENABLED:
        if not FUpdatingEnabled then
        begin
          FUpdatingEnabled := True;
          try
            Enabled := GetWordBoolProp(DISPID_ENABLED);
          finally
            FUpdatingEnabled := False;
          end;
        end;
      DISPID_FONT:
        if not FUpdatingFont then
        begin
          FUpdatingFont := True;
          try
            OleFontToFont(GetVariantProp(DISPID_FONT), Font);
          finally
            FUpdatingFont := False;
          end;
        end;
      DISPID_FORECOLOR:
        if not FUpdatingFont then
        begin
          FUpdatingFont := True;
          try
            Font.Color := GetIntegerProp(DISPID_FORECOLOR);
          finally
            FUpdatingFont := False;
          end;
        end;
    end;
  except  // control sent us a notification for a dispid it doesn't have.
    on EOleError do ;
  end;
  Result := S_OK;
end;

function TpsvApplet.OnRequestEdit(dispid: TDispID): HResult;
begin
  Result := S_OK;
end;

{ TpsvApplet.ISimpleFrameSite }

function TpsvApplet.PreMessageFilter(wnd: HWnd; msg, wp, lp: Integer;
  out res: Integer; out Cookie: Longint): HResult;
begin
  Result := S_OK;
end;

function TpsvApplet.PostMessageFilter(wnd: HWnd; msg, wp, lp: Integer;
  out res: Integer; Cookie: Longint): HResult;
begin
  Result := S_OK;
end;


procedure TpsvApplet.InitControlData;
const
  CControlData: TControlData2 =(
    LicenseKey: nil) ;
begin
  CControlData.ClassID := FHost.FGUID;
  ControlData := @CControlData;
end;

function FontToOleFont(Font: TFont): Variant;
var
  Temp: IFontDisp;
begin
  GetOleFont(Font, Temp);
  Result := Temp;
end;

procedure OleFontToFont(const OleFont: Variant; Font: TFont);
begin
  SetOleFont(Font, IUnknown(OleFont) as IFontDisp);
end;

constructor TpsvServerOpt.Create(AOwner : TpsvAppletHost);
begin
  FOwner        := AOwner;
  FAutoTransfer := False;
  FServerFolder := EmptyStr;
  FServerName   := EmptyStr;
  FUserName     := EmptyStr;
  FPassword     := EmptyStr;
  FproxyServer  := EmptyStr;
  FConnectType  := ctDirect;
end;

procedure   TpsvServerOpt.Assign(Source : TPersistent);
begin
  if Source is TpsvServerOpt then
    begin
     FUserName := TpsvServerOpt(Source).UserName;
     FPassword := TpsvServerOpt(Source).Password;
     FServerFolder := TpsvServerOpt(Source).ServerFolder;
     FAutoTransfer := TpsvServerOpt(Source).AutoTransfer;
     FTransferType := TpsvServerOpt(Source).TransferType;
     FConnectType  := TpsvServerOpt(Source).ConnectType;
     FProxyServer := TpsvServerOpt(Source).ProxyServer;
     inherited Assign(Source);
   end;
end;

constructor TpsvFileOpt.Create(AOwner : TpsvAppletHost);
begin
  FOwner := AOwner;
  FDefaultExt := 'ocx';
  FFolder     := 'C:\';
  FFileName   := EmptyStr;
end;

procedure TpsvFileOpt.SetDefaultExt(Value : string);
begin
  if Value <> FDefaultExt then
    begin
      if Value[1] = '.' then
        Delete(Value,1,1);
        FDefaultExt := Value;
    end;
end;

procedure TpsvFileOpt.SetFileName(Value : String);
begin
  if not (csLoading in FOwner.ComponentState) then
   if FOwner.Active then
    raise Exception.Create('Cannot change file name while host is active');
  if Value = EmptyStr then Exit;


  if (Pos('.',Value) = 0) then
    Value := Value + '.'+ FDefaultExt;
  FFileName := Value;

  if FileExists(FFolder+Value) then
   begin
     If FOwner.AutoExtractGUID then
       begin
         FOwner.FGUID    := FOwner.ExtractGUID;
         FOwner.FGUIDStr := GUIDToString(FOwner.FGUID);
       end;
   end;
end;

procedure TpsvFileOpt.SetFolder(Value : string);
begin
  If Value[Length(Value)] <> '\' then
    Value := Value + '\';
  FFolder := Value;
end;

procedure   TpsvFileOpt.Assign(Source : TPersistent);
begin
  if Source is TpsvFileOpt then
    begin
     FFileName := TpsvFileOpt(Source).FileName;
     FFolder   := TpsvFileOpt(Source).Folder;
     FDefaultExt := TpsvFileOpt(Source).DefaultExt;
     inherited Assign(Source);
   end;
end;


procedure Register;
begin
  RegisterComponents('System',[TpsvAppletHost]);
end;

end.
